package net.hoo2.auth.dsproject.snake;

import java.util.*;

/**
 * @class HeuristicPlayer
 * @brief Represent a Heuristic Player in the Game
 * 
 * The players are playing in a round-robin sequence and we keep track
 * for each one of them their playing order, score and place on the board.
 * This kind of player, is a cheater. He can control the dice. Not fair dude.
 * 
 * @author Christos Choutouridis AEM:8997
 * @email  cchoutou@ece.auth.gr
 */
public class HeuristicPlayer
             extends Player {
   // evaluate configuration
   static final double FACTOR_TILE      = 0.15;
   static final double FACTOR_FWD_STEP  = 0.55;
   static final double FACTOR_BACK_STEP = -1.0;
   static final double FACTOR_UP_POINTS = 0.3;
   static final double FACTOR_DWN_POINTS= -1.0;
   static final double FACTOR_INIT_EVAL = 1.0;
   static final double FACTOR_ROUND     = -1.0;
   static final double FACTOR_GAME      = 10.0;

   /** @name Constructors */
   /** @{ */
   /** Default doing nothing constructor */
   public HeuristicPlayer() {
      super ();
      path     = new ArrayList<Integer[]>();
      evalData = new EvalData[board.getN() * board.getM() +1];
   }
   /**
    * @brief   The main constructor
    * 
    * This creates a player for the game
    * @param playerId   The player's to create
    * @param name       The name of the player
    * @param board      Reference to the board the player will play on.
    */
   HeuristicPlayer (int playerId, String name, Board board) {
      super (playerId, name, board);
      path     = new ArrayList<Integer[]>();
      evalData = new EvalData[board.getN() * board.getM() +1];
   }
   /* @} */

   /** @name Get/Set interface */
   /** @{ */
   ArrayList<Integer[]> getPath() { return path; }
   void setPath (ArrayList<Integer[]> path) {
      this.path = path;
   }
   /** @} */
   
   /**
    * Override dice functionality for the player
    * @return  As this is called from the game only to select playing order 
    *          we cheat and return 1
    */
   @Override
   int dice () {
      return 1;
   }
   
   /**
    * Override get the next tile after the user's move
    * @param tile    The initial tile
    * @return  The tile after the move
    */
   @Override
   int getNextMove (int tile) {
      int   roll = 0;

      // Evaluate each possible dice result and find the better one
      roll = evalProcess(tile);
      // Do the move and get the move data
      Integer[] move_data = Arrays.stream(move (tile, roll, true))
                                  .boxed()
                                  .toArray(Integer[]::new);
      // Store the move data
      path.add(move_data);
      return tile + roll;  // return the new tile position
   }

   /**
    * The Heuristic statistics version
    * @param verbose    Flag to select the verbosity
    * @param sum        Flag to select if we need to print a summarize of the user hystory
    */
   @Override
   void statistics (boolean verbose, boolean sum) {
      if (sum) {
         // If we run the summarize
         int nSnakes      =0;
         int nLadders     =0;
         int nRedApples   =0;
         int nBlackApples =0;

         // Calculate frequencies
         for (int i=0 ; i<path.size() ; ++i) {
            nSnakes += path.get(i)[MOVE_SNAKES_IDX];
            nLadders+= path.get(i)[MOVE_LADDERS_IDX];
            nRedApples   += path.get(i)[MOVE_RED_APPLES_IDX];
            nBlackApples += path.get(i)[MOVE_BLACK_APPLES_IDX];
         }
         // Print the results
         System.out.println("");
         System.out.println("*** Statistics for " + name + " ***");
         System.out.println("   Number of Snake bites       : " + nSnakes);
         System.out.println("   Number of Ladders used      : " + nLadders);
         System.out.println("   Number of Red Apples eaten  : " + nRedApples);
         System.out.println("   Number of Black Apples eaten: " + nBlackApples);
         
      }
      else
         // Call the base version
         super.statistics(verbose, sum);
   }

   int evalProcess (int tile) {
      int[] check = new int[MOVE_DATA_SIZE];
      int begin;
      int end = board.getN() * board.getM();
      int roll;
      EvalData e = new EvalData();

      for (int i =0 ; i<=end ; ++i) {
         evalData[i] = new EvalData();
         evalData[i].evaluation = Double.NEGATIVE_INFINITY;
      }
      for (begin =tile ; begin < end ; ++begin) {
         if ((board.checkLadder(begin, false) != begin) ||
             (board.checkSnake(begin) != begin))
            continue;
         int initRound   = evalData[begin].round;
         double initEval = evalData[begin].evaluation;
         for (roll = 1 ; roll <= 6 ; ++roll) {
            check = move (begin, roll, false);
            e.game = _saturate (check);
            e.round = initRound +1;
            e.from = begin;
            e.roll = roll;
            e.steps = check[MOVE_STEPS_IDX];
            e.points = check[MOVE_POINTS_IDX];
            e.evaluation = _evalFormula (check[MOVE_TILE_IDX], e.steps, e.points, e.round, initEval, e.game);
            if (e.evaluation > evalData[check[MOVE_TILE_IDX]].evaluation) {
               evalData[check[MOVE_TILE_IDX]].copy(e);
            }
         }
      }
      // find route back of the optimal path
      e = evalData[board.getN() * board.getM()];
      EvalData last = e;
      while (e.from != tile) {
         last = e;
         e = evalData[e.from];
      } 
      
      return last.roll;    // return the first choice
   }

   double evaluate (int tile, int roll) {
      return 0;
   }

   /**
    * The main evaluation function
    * @param tile    The current tile of the player
    * @param roll    the roll to check
    * @return        The evaluation of the roll
    */
   private double _evalFormula (int tile, int steps, int points, int round, double initEval, boolean game) {
      initEval = (initEval == Double.NEGATIVE_INFINITY) ? 0 : initEval;
      return     tile * FACTOR_TILE +
                steps * ((steps>0)  ? FACTOR_FWD_STEP : FACTOR_BACK_STEP) +
               points * ((points>0) ? FACTOR_UP_POINTS : FACTOR_DWN_POINTS) +
             initEval * FACTOR_INIT_EVAL +
                round * FACTOR_ROUND + 
       ((game) ? 1:0) * FACTOR_GAME;
   }

   private boolean _saturate (int[] check) {
      // make adjustments if game is finished
      int tiles    = board.getM() * board.getN();
      boolean game = (check[MOVE_TILE_IDX] >= tiles) ? true : false;

      if (check[MOVE_TILE_IDX]> tiles) {
         check[MOVE_STEPS_IDX] -= check[MOVE_TILE_IDX] - tiles;
         check[MOVE_TILE_IDX]   = tiles;
      }
      return game;
   }

   /** @name Data members package access only */
   /** @{ */
   private ArrayList<Integer[]> path;  /**< Players history as required */
   private EvalData[] evalData; 
   /** @} */
}


class EvalData {
   int     round;
   int     from;
   int     roll;
   int     steps;
   int     points;
   boolean game;
   double  evaluation;
   EvalData () { }
   
   void copy (EvalData e) {
      round       = e.round;
      from        = e.from;
      roll        = e.roll;
      steps       = e.steps;
      points      = e.points;
      game        = e.game;
      evaluation  = e.evaluation;
   }
}